#include <math.h>
#include <stdio.h>
#include <errno.h>
#include <limits.h>
#include "utils.h"
#include "config.h"
#include "asim.h"

#define INACTIVE 0
#define POTENTIALLY_ACTIVE 1
#define ACTIVE 2

#define NAT_LOG_E 2.7182818

#define UPDATE(fmt, ...)                                             \
  do {									\
    if (VFLAG_UPDATE & _config.verbose_flags)                           \
      fprintf(_update_file, "%d " fmt "\n", _sim_iters, __VA_ARGS__);	\
  } while (0);

CREATE_TIMER(MAIN_TIMER);
CREATE_TIMER(INIT_TIMER);

// Globals. Always prefixed with _
FILE* _update_file;       // the file to write updates to

void init_locs();
int get_pop(int i);
void sort_locs_by_pop();
int cmp_int(const void* a, const void* b);
void init_agents();
void add_agent();
void add_to_loc(shared agent_t* agent, shared loc_t* loc);
shared agent_t* add_link_at_loc(shared agent_t* agent, shared loc_t* loc, int is_new);
int agents_linked(shared agent_t* agent_1, shared agent_t* agent_2);
void link_agents(shared agent_t* agent_1, shared agent_t* agent_2);
void remove_link(shared agent_t* agent, int link_id);
void remove_link();
void add_links();
void extend_agents();
int get_desired_extent(shared agent_t* agent);
void update_agents();
void do_traffic();
uint64_t get_traffic_volume(int src, int dest);
uint64_t get_int_traffic_volume(int i);
void compute_graph_paths();
void compute_aprobs();
int get_rnd_loc();
int get_occupied_rnd_loc();
int get_agent_new_rnd_loc(shared agent_t* agent);
int get_rnd_agent_at_loc(shared agent_t* agent, shared loc_t* loc);
uint64_t get_traffic_from_deg(int deg);
double get_wickedness();
void print_stats(double start_t, double* prev_t);

int main(int argc, char** argv) {
  clear_timers();
  SERIAL {
     printf("# ASIM (Version %.2f) compiled %s\n", ASIM_VERSION, __DATE__);
    load_config(argc, argv, &_config);
    _update_file = fopen("updates.dat", "w");
    print_config(&_config, '#', _update_file);
  }
  upc_barrier;
  timer_start(INIT_TIMER);
  init_rnd(_config.rseed + MYTHREAD);
  init_locs();
  init_agents();
  upc_barrier;
  int restored = 0;
  SERIAL {
    if (_config.snapshot_fname) {
      restore_snapshot();
      restored = 1;
    }
  }
  timer_stop(INIT_TIMER);
  upc_barrier;
  SERIAL {
    printf("# Initialization time %.2f\n", timer_read(INIT_TIMER));
    time_t tt = time(NULL);
    printf("# %s", ctime(&tt));
    _sim_iters = _agents.num;
    if (_config.output_interval) {
      mprintf(13, "%6s", "# n", 
	      "%7s", "av k", 
	      "%6s", "mx k", 
	      "%7s", "cover",
	      "%8s", "av sz", 
	      "%8s", "mx sz", 
	      "%10s", "av pop", 
	      "%10s", "traf", 
	      "%10s", "etraf", 
	      "%8s", "btraf", 
	      "%8s", "lgtraf",
	      "%8s", "wckdns", 
	      "%8s", "acts");
      if (!_config.snapshot_fname) printf("%10s", "ETA");
      printf("\n");
    }
    fflush(stdout);
  }
  upc_barrier;
  timer_start(MAIN_TIMER);
  double start_t = _get_current_time();
  double prev_t = _get_current_time() - start_t;
  char buf[100];
  while (_agents.num < _config.max_agents) {
    upc_barrier;
    PHASE(add_agent);
    PHASE(add_links);
    PHASE(extend_agents);
    PHASE(update_agents);
    PHASE(compute_graph_paths);
    PHASE(do_traffic);
    SERIAL {
      _sim_iters++;
      if ((_config.output_interval && !(_sim_iters % _config.output_interval)) ||
	  _agents.num == _config.max_agents) {
	print_stats(start_t, &prev_t);
	if (_config.verbose_flags & VFLAG_DISTS) print_dists();
	if (_config.verbose_flags & VFLAG_AGENTS) print_agents();
      }
    }
    if (_agents._[0].id != 0) FAIL("agent %d has id %d\n", 0, _agents._[0].id);
  } 
  upc_barrier;
  timer_stop(MAIN_TIMER);
  double elapsed_main_time = timer_read(MAIN_TIMER);
  if (_config.verbose_flags & VFLAG_TIMERS) print_timers();
  double verify_time = verify();
  upc_barrier;
  if (_config.verbose_flags & VFLAG_APROBS) compute_aprobs();
  upc_barrier;
  SERIAL {
    printf("# Ran %d iterations in %.2fs (%.2fs total): %d agents, %.3f av deg, "
           "%.2f cover, %.2f av size\n",
           _sim_iters, elapsed_main_time, elapsed_main_time + timer_read(INIT_TIMER) + verify_time,
	   _agents.num, get_av_deg(), get_locs_covered(), get_av_agent_size());
    if (_config.verbose_flags & VFLAG_AGENTS) print_agents();
    if (_config.verbose_flags & VFLAG_LOCS) print_locs();
    if (_config.verbose_flags & VFLAG_LINKS) print_links();
    if (_config.verbose_flags & VFLAG_DISTS) print_dists();
    if (_config.verbose_flags & VFLAG_UPDATE) fclose(_update_file);
    save_snapshot();
  }
  return 0;
}

void init_locs() {
  SERIAL {
    printf("# Initializing locations ...");
    fflush(stdout);
    _locs.num = NUM_LOCS;
    _locs.num_occupied = 0;
    _locs.tot_pop = 0;
    _locs.occupied_pop = 0;
  }
  upc_barrier;
  shared int* min_pops;
  shared int* max_pops;
  shared int* tot_pops;
  upc_barrier;
  _locs._ = upc_all_alloc(NUM_LOCS, sizeof(loc_t));
  min_pops = upc_all_alloc(THREADS, sizeof(int));
  max_pops = upc_all_alloc(THREADS, sizeof(int));
  tot_pops = upc_all_alloc(THREADS, sizeof(int));
  upc_barrier;
  min_pops[MYTHREAD] = MAX_POP;
  max_pops[MYTHREAD] = 0;
  tot_pops[MYTHREAD] = 0;
  upc_forall (int i = 0; i < _locs.num; i++; i) {
    _locs._[i].id = i;
    _locs._[i].num_agents = 0;
    _locs._[i].pop = get_pop(i);
    _locs._[i].agent_list = upc_global_alloc(_config.max_agents, sizeof(unsigned short));
    tot_pops[MYTHREAD] += _locs._[i].pop;
    if (max_pops[MYTHREAD] < _locs._[i].pop) max_pops[MYTHREAD] = _locs._[i].pop;
    if (min_pops[MYTHREAD] > _locs._[i].pop) min_pops[MYTHREAD] = _locs._[i].pop;
  }
  upc_barrier;
  upc_forall (int i = 0; i < _locs.num; i++; i) {
    if (_locs._[i].id != i) FAIL("loc %d has id %d\n", i, _locs._[i].id);
  }
  upc_barrier;
  SERIAL {
    _locs.max_pop = 0;
    _locs.min_pop = MAX_POP;
    _locs.tot_pop = 0;
    for (int i = 0; i < THREADS; i++) {
      if (max_pops[i] > _locs.max_pop) _locs.max_pop = max_pops[i];
      if (min_pops[i] < _locs.min_pop) _locs.min_pop = min_pops[i];
      _locs.tot_pop += tot_pops[i];
    }
    printf(" max pop %d, min pop %d, total pop %d\n", _locs.max_pop, _locs.min_pop, _locs.tot_pop);
  }
  upc_barrier;
  // sorting the locations improves the weighted random selection speed
  SERIAL {
    sort_locs_by_pop();
    print_locs();
  }
  upc_barrier;
}

int get_pop(int i) {
  double min_p = pow(MAX_POP, 1.0 / _config.pop_distr_exp);
  double pop;
  do {
    pop = pow(get_double_rnd(min_p, 1.0), _config.pop_distr_exp);
  } while (pop > MAX_POP * 0.8);
  if (pop < 1) pop = 1;
  if (pop > MAX_POP) pop = MAX_POP;
  return pop;
}

void sort_locs_by_pop() {
  // qsort in serial is fast enough to do this once for a smallish array
  SERIAL {
    int* loc_pops = calloc(_locs.num, sizeof(int));
    for (int i = 0; i < _locs.num; i++) loc_pops[i] = _locs._[i].pop;
    qsort(loc_pops, _locs.num, sizeof(int), cmp_int);
    for (int i = 0; i < _locs.num; i++) _locs._[i].pop = loc_pops[i];
    free(loc_pops);
  }
}

int cmp_int(const void* a, const void* b) {
  int ai = *(int *)a;
  int bi = *(int *)b;
  return (ai < bi ? 1 : (ai > bi ? -1 : 0));
}

void init_agents() {
  SERIAL {
    _agents.num = 0;
    _agents.num_links = 0;
    _agents.max_deg = 0;
    _agents.num_actives = 0;
  }
  upc_barrier;
  _agents._ = upc_all_alloc(_config.max_agents, sizeof(agent_t));
  upc_barrier;
  upc_forall(int i = 0; i < _config.max_agents; i++; i) {
    shared agent_t* agent = &_agents._[i];
    agent->id = i;
    agent->tot_pop = 0;
    agent->pop_sh = 0;
    agent->funds = _config.base_income;
    agent->traffic = 0;
    agent->lost_good_traffic = 0;
    agent->lost_bad_traffic = 0;
    agent->end_user_traffic = 0;
    agent->bad_traffic = 0;
    agent->num_links = 0;
    agent->num_locs = 0;
    agent->birth_time = 0;
    agent->iters = 0;
    agent->av_dist = 0;
    agent->max_dist = 0;
    agent->selection_count = 0;
    agent->is_stub = 0;
    //    if (get_double_rnd(0, 1.0) < 0.3) agent->is_stub = 1;
    agent->wickedness = get_wickedness();
    agent->active_intrvns = INACTIVE;
    // although this is set here, interventions will only happen when (if) the agent gets large enough 
    if (get_double_rnd(0, 1) < _config.intrvn_active_prob) agent->active_intrvns = POTENTIALLY_ACTIVE;
    agent->links = upc_global_alloc(_config.max_agents, sizeof(unsigned short));
    agent->links_list = upc_global_alloc(_config.max_agents, sizeof(unsigned short));
    agent->locs = upc_global_alloc(NUM_LOCS, sizeof(unsigned short));
    agent->locs_list = upc_global_alloc(NUM_LOCS, sizeof(unsigned short));
    agent->dists = upc_global_alloc(_config.max_agents, sizeof(unsigned short));
    agent->paths = upc_global_alloc(_config.max_agents, sizeof(unsigned short));
    agent->dist_q = upc_global_alloc(1, sizeof(queue_t));
    if (!init_queue(agent->dist_q, _config.max_agents)) FAIL("Couldn't init q for agent %d\n", i);
    for (int l = 0; l < NUM_LOCS; l++) agent->locs[l] = 0;
    for (int l = 0; l < _config.max_agents; l++) agent->links[l] = 0;
  }
  upc_barrier;
  upc_forall(int i = 0; i < _config.max_agents; i++; i) {
    // sanity check
    if (_agents._[i].id != i) FAIL("agent %d has id %d\n", i, _agents._[i].id);
  }
  upc_barrier;
}

double get_wickedness() {
  double w = -_config.wickedness * log(1.0 - get_double_rnd(0, 1.0));
  if (w > 0.5) w = 0.5;
  return w;
}

void add_agent() {
  SERIAL {
    int agent_i = _agents.num;
    _agents.num++;
    shared agent_t* agent = get_agent_ptr(agent_i, __LINE__);
    int loc_i = -1;
    // always start at the most populous location, to take away some of the variability betw runs
    if (agent_i == 0) loc_i = 0;//get_rnd_loc();
    else loc_i = get_occupied_rnd_loc();
    if (loc_i < 0) FAIL("Cannot get location for agent %d\n", agent_i);
    shared loc_t* loc = get_loc_ptr(loc_i);
    agent->birth_time = _sim_iters;
    UPDATE("mkAS %d", agent->id);
    DBG(VFLAG_ADD_AGENT, "Adding agent %d %d\n", agent->id, agent_i);
    add_to_loc(agent, loc);
    if (agent->id > 0) {
      shared agent_t* neighbor = add_link_at_loc(agent, loc, 1);
      if (!neighbor) {
	FAIL("Cannot find a neighbor in %d others for new agent %d at location %d. "
	     "This shouldn't happen.\n", loc->num_agents, agent->id, loc->id);
      }
    }
  }
}

int get_rnd_loc() {
  int rnd_pos = get_int_rnd(_locs.tot_pop);
  for (int i = 0; i < _locs.num; i++) { 
    rnd_pos -= _locs._[i].pop;
    if (rnd_pos < 0) return i; 
  } 
  return -1; 
}

int get_occupied_rnd_loc() {
  int rnd_pos = get_int_rnd(_locs.occupied_pop);
  for (int i = 0; i < _locs.num; i++) { 
    shared loc_t* loc = get_loc_ptr(i);
    if (loc->num_agents) rnd_pos -= loc->pop;
    if (rnd_pos < 0) return i; 
  } 
  return -1; 
}

void add_to_loc(shared agent_t* agent, shared loc_t* loc) {
  DBG(VFLAG_EXTEND, "Extending agent %d to loc %d, pop %d\n", agent->id, loc->id, loc->pop);
  agent->locs[loc->id] = 1;
  agent->locs_list[agent->num_locs] = loc->id;
  agent->num_locs++;
  if (agent->num_locs > _locs.num) {
    FAIL("agent %d has too many locs %d\n", agent->id, agent->num_locs);
  }
  agent->tot_pop += loc->pop;
  int num_agents = bupc_atomicI32_fetchadd_relaxed(&loc->num_agents, 1);
  if (num_agents > _agents.num) FAIL("loc %d has too many agents %d\n", loc->id, num_agents);
  loc->agent_list[num_agents] = agent->id;
  if (!num_agents) {
    bupc_atomicI32_fetchadd_relaxed(&_locs.occupied_pop, loc->pop);
    bupc_atomicI32_fetchadd_relaxed(&_locs.num_occupied, 1);
  }
}

shared agent_t* add_link_at_loc(shared agent_t* agent, shared loc_t* loc, int is_new) {
  int dest_agent_i = get_rnd_agent_at_loc(agent, loc);
  if (dest_agent_i == -1) return NULL;
  if (dest_agent_i >= _agents.num) {
    FAIL("Trying to link to agent %d which is not yet in existence\n", dest_agent_i);
  }
  shared agent_t* dest_agent = get_agent_ptr(dest_agent_i, __LINE__);
  if (dest_agent->is_stub && !is_new) return NULL;
  DBG(VFLAG_ADD_LINK, "Linking agent %d to agent %d at loc %d\n", 
      agent->id, dest_agent->id, loc->id);
  link_agents(dest_agent, agent);
  link_agents(agent, dest_agent);
  return dest_agent;
}

int get_rnd_agent_at_loc(shared agent_t* agent, shared loc_t* loc) { 
  if (loc->num_agents - 1 <= 0) return -1; 
  int rnd_pos = get_int_rnd(loc->num_agents - 1);
  for (int i = 0; i < loc->num_agents; i++) { 
    int dest_id = loc->agent_list[i];
    if (dest_id >= _agents.num) FAIL("Trying to get agent at dest_id %d\n", dest_id);
    if (agent->id != dest_id && !agent->links[dest_id]) rnd_pos--;
    if (rnd_pos < 0) return dest_id;
  } 
  return -1; 
}


void link_agents(shared agent_t* agent_1, shared agent_t* agent_2) {
  if (agent_1->num_links >= MAX_DEG) {
    FAIL("Agent %d has too many links, %d > %d\n", agent_1->id, agent_1->num_links, MAX_DEG);
  }
  if (agent_1->num_links >= _agents.num) {
    FAIL("agent %d has more links (%d) than there are agents\n", agent_1->id, agent_1->num_links);
  }
  if (agent_2->id >= _agents.num) FAIL("Agent %d out of range\n", agent_2->id);
  agent_1->links[agent_2->id] = 1;
  agent_1->links_list[agent_1->num_links] = agent_2->id;
  agent_1->num_links++;
  _agents.num_links++;
  UPDATE("mkASedge %d %d", agent_1->id, agent_2->id);
}

void remove_links() {
  if (_agents.num < 2) return;
  for (int i = 0; i < 100; i++) {
    shared agent_t* agent1 = get_agent_ptr(get_int_rnd(_agents.num), __LINE__);
    if (agent1->num_links < 2) continue;
    shared agent_t* agent2 = get_agent_ptr(agent1->links_list[get_int_rnd(agent1->num_links)], __LINE__);
    if (agent2->num_links < 2) continue;
    DBG(VFLAG_ADD_LINK, "removing link from %d (%d) to %d (%d)\n", 
	   agent1->id, agent1->num_links, agent2->id, agent2->num_links);
    remove_link(agent1, agent2->id);
    remove_link(agent2, agent1->id);
    _agents.num_links -= 2;
    break;
  } 
}

void remove_link(shared agent_t* agent, int link_id) {
  int link_pos = 0;
  for (int i = 0; i < agent->num_links; i++) {
    if (agent->links_list[i] == link_id) {
      link_pos = i;
      break;
    }
  }
  for (int i = link_pos; i < agent->num_links - 1; i++) agent->links_list[i] = agent->links_list[i + 1];
  agent->num_links--;
  agent->links[link_id] = 0;
  DBG(VFLAG_ADD_LINK, "agent %d has link to %d removed, now has %d links\n", 
      agent->id, link_id, agent->num_links);
}

void add_links() {
  SERIAL {
    //    if (REMOVE_LINKS) remove_links();
    shared loc_t* loc = NULL;
    shared agent_t* agent = NULL;
    int links_needed = _config.av_degree * _agents.num - _agents.num_links;
    if (links_needed <= 0) return;
    DBG(VFLAG_ADD_LINK, "Trying to add %d new links\n", links_needed);
    int num_samples = 10;
    switch (LINK_STRATEGY) {
    case SPREAD_GEO_SEL_LSTRAT:
      // choose agents at random according to geog. selection
      do {
	loc = get_loc_ptr(get_occupied_rnd_loc());
	if (loc->num_agents == 1) continue;
	agent = get_agent_ptr(loc->agent_list[get_int_rnd(loc->num_agents)], __LINE__);
	if (add_link_at_loc(agent, loc, 0)) links_needed--;
      } while (--num_samples && links_needed);
      break;
    case FOCUS_GEO_SEL_LSTRAT:
      // choose one agent and try to add all links to that single agent
      do {
	loc = get_loc_ptr(get_occupied_rnd_loc());
      } while (--num_samples && loc->num_agents == 1);
      agent = get_agent_ptr(loc->agent_list[get_int_rnd(loc->num_agents)], __LINE__);
      num_samples = 10;
      do {
	loc = get_loc_ptr(get_occupied_rnd_loc());
	if (!agent->locs[loc->id]) add_to_loc(agent, loc);
	if (add_link_at_loc(agent, loc, 0)) links_needed--;
      } while (--num_samples && links_needed);
      break;
    case SPREAD_URND_SRC_LSTRAT:
      // choose link initiator at random, and target with geog. sel. 
      // will also extend to new loc as needed
      // this works well, except for CC and K-cores.
      do {
	agent = get_agent_ptr(get_int_rnd(_agents.num), __LINE__);
	loc = get_loc_ptr(get_occupied_rnd_loc());
	if (!agent->locs[loc->id]) add_to_loc(agent, loc);
	if (add_link_at_loc(agent, loc, 0)) links_needed--;
      } while (--num_samples && links_needed);
      break;
    case FOCUS_URND_SRC_LSTRAT:
      // choose link initiator at random, and target with geog. sel. 
      // try to keep one original node
      agent = get_agent_ptr(get_int_rnd(_agents.num), __LINE__);
      num_samples = 10;
      do {
	loc = get_loc_ptr(get_occupied_rnd_loc());
	if (!agent->locs[loc->id]) add_to_loc(agent, loc);
	if (add_link_at_loc(agent, loc, 0)) links_needed--;
      } while (--num_samples && links_needed);
      break;
    case CHAIN_LSTRAT:
      // choose starting agent at uniform rnd from all agents, choose target with geo sel
      // each agent that receives a link tries to link in turn, using geo sel
      // Works the best overall. The biggest problem is that the path length is too short.
      // This is in conjunction with limiting the base_income to the first 100 agents and 
      // setting num_samples = 5
      // If we set the num agents with base_income to all of them, then the path length is about
      // correct but other values are worse.
      num_samples = 5;
      do {
	if (!agent) agent = get_agent_ptr(get_int_rnd(_agents.num), __LINE__);
	if (agent->is_stub && agent->num_links >= 2) continue;
	loc = get_loc_ptr(get_occupied_rnd_loc());
	if (!agent->locs[loc->id]) {
	  add_to_loc(agent, loc);
	  agent->funds -= _config.ext_cost;
	}
	agent = add_link_at_loc(agent, loc, 0);
	if (agent) links_needed--;
      } while (--num_samples && links_needed);
      break;
    case URND_FIXED_GEO_LSTRAT:
      // choose an agent at random, then chose one of its already occupied locations, and 
      // add a link there if possible. Try do this multiple times per single agent
      agent = get_agent_ptr(get_int_rnd(_agents.num), __LINE__);
      num_samples = 100;
      do {
	loc = get_loc_ptr(agent->locs_list[get_int_rnd(agent->num_locs)]);
	if (add_link_at_loc(agent, loc, 0)) links_needed--;
      } while (--num_samples && links_needed);
      break;
    case GEO_SEL_FIXED_GEO_LSTRAT:
      // choose an agent according to geo sel, then chose one of its already occupied locations, and 
      // add a link there if possible. Try do this multiple times per single agent
      loc = get_loc_ptr(get_occupied_rnd_loc());
      agent = get_agent_ptr(loc->agent_list[get_int_rnd(loc->num_agents)], __LINE__);
      num_samples = 10;
      do {
	loc = get_loc_ptr(agent->locs_list[get_int_rnd(agent->num_locs)]);
	if (add_link_at_loc(agent, loc, 0)) links_needed--;
      } while (--num_samples && links_needed);
      break;
    }
  }
}

void extend_agents() {
  if (_sim_iters % GRAPH_EVAL_FREQ) return;
  upc_forall (int i = 0; i < _agents.num; i++; i) {
    shared agent_t* agent = get_agent_ptr(i, __LINE__);
    //    int desired_extent = get_desired_extent(agent);
    //    if (desired_extent >= _locs.num) {
    //      continue;
      //      FAIL("Agent %d is trying to expand to beyond available loc size %d. "
      //	   "Rerun with more locations\n", agent->id, _locs.num);
    //    }
    //    int locs_needed = desired_extent - agent->num_locs;
    if (agent->funds < _config.ext_cost) continue;
    if (agent->num_locs == _locs.num) continue;
    int locs_needed = agent->funds / _config.ext_cost;
    if (locs_needed + agent->num_locs >= _locs.num) locs_needed = _locs.num - agent->num_locs;
    agent->funds -= (locs_needed * _config.ext_cost);
    for (int extent_i = 0; extent_i < locs_needed; extent_i++) {
      int loc_i = get_agent_new_rnd_loc(agent);
      if (loc_i == -1) {
	FAIL("Cannot get location for extending agent %d from %d to %d locs\n", 
	     agent->id, agent->num_locs, locs_needed);
      }
      add_to_loc(agent, get_loc_ptr(loc_i));
    }
  }
}

int get_desired_extent(shared agent_t* agent) {
  double income = agent->traffic;
  if (_config.full_traffic) income /= ((double)MAX_POP * (double)MAX_POP);
  if (_agents.num < NUM_AGENTS_WITH_BASE_INCOME) income += _config.base_income;
  double extent = income / _config.ext_cost;
  if (get_double_rnd(0, 1.0) <= extent - trunc(extent)) extent++;
  if (extent < 1) extent = 1;
  if (agent->is_stub && extent > 2) extent = 2;
  return extent;
}

int get_agent_new_rnd_loc(shared agent_t* agent) {
  int rnd_pos = get_int_rnd(_locs.tot_pop - agent->tot_pop);
  for (int i = 0; i < _locs.num; i++) { 
    if (!agent->locs[i]) rnd_pos -= _locs._[i].pop;
    if (rnd_pos < 0) return i; 
  } 
  return -1; 
}

void update_agents() {
  shared int* max_degs = upc_all_alloc(THREADS, sizeof(int));
  shared double* tot_wickedness = upc_all_alloc(THREADS, sizeof(double));
  upc_barrier;
  tot_wickedness[MYTHREAD] = 0;
  max_degs[MYTHREAD] = 0;
  upc_forall (int i = 0; i < _agents.num; i++; i) {
    shared agent_t* agent = get_agent_ptr(i, __LINE__);
    agent->iters++;
    agent->pop_sh = 0;
    if (!(_sim_iters % GRAPH_EVAL_FREQ)) {
      if (_config.full_traffic) {
	agent->funds += (double)agent->traffic / ((double)MAX_POP * (double)MAX_POP);
      } else agent->funds += agent->traffic;
    }
    // sanity check
    if (agent->id != i) FAIL("agent %d has id %d\n", i, agent->id);
    for (int l = 0; l < agent->num_locs; l++) {
      shared loc_t* loc = get_loc_ptr(agent->locs_list[l]);
      agent->pop_sh += ((double)loc->pop / (double)loc->num_agents);
    }
    if (agent->num_links > max_degs[MYTHREAD]) max_degs[MYTHREAD] = agent->num_links;
    agent->wickedness -= _config.recovery_rate;
    double bad_frac = 0;
    if (agent->end_user_traffic - agent->lost_good_traffic - agent->lost_bad_traffic > 0) {
      bad_frac = agent->bad_traffic / (agent->end_user_traffic - agent->lost_good_traffic - agent->lost_bad_traffic);
    }
    agent->wickedness +=  (_config.infct_rate * bad_frac);
    if (agent->wickedness > 0.5) agent->wickedness = 0.5;
    if (agent->wickedness < 0) agent->wickedness = 0;
    tot_wickedness[MYTHREAD] += agent->wickedness;
    if (agent->active_intrvns == POTENTIALLY_ACTIVE && agent->num_links >= _config.intrvn_large_isp_deg) {
      int32_t num_actives = bupc_atomicI32_fetchadd_relaxed(&_agents.num_actives, 1);
      if (num_actives >= 20 && _config.intrvn_active_prob == 1) {
	bupc_atomicI32_fetchadd_relaxed(&_agents.num_actives, -1);
	agent->active_intrvns = INACTIVE;
      } else agent->active_intrvns = ACTIVE;
    }
  }
  upc_barrier;
  SERIAL {
    _agents.max_deg = 0;
    _agents.av_wickedness = 0;
    for (int i = 0; i < THREADS; i++) {
      if (max_degs[i] > _agents.max_deg) _agents.max_deg = max_degs[i];
      _agents.av_wickedness += tot_wickedness[i];
    }
    _agents.av_wickedness /= _agents.num;
    upc_free(max_degs);
    upc_free(tot_wickedness);
  }
}

void do_traffic() {
  // the gravity model
  if ((_sim_iters % GRAPH_EVAL_FREQ) && _agents.num < _config.max_agents) return;
  // need to start with zero traffic for all of them
  upc_forall (int i = 0; i < _agents.num; i++; i) {
    _agents._[i].traffic = 0;
    _agents._[i].bad_traffic = 0;
    _agents._[i].end_user_traffic = 0;
    _agents._[i].lost_good_traffic = 0;
    _agents._[i].lost_bad_traffic = 0;
  }
  upc_barrier;
  upc_forall (int i = 0; i < _agents.num; i++; i) {
    for (int dest = 0; dest < _agents.num; dest++) {
      int src = i;
      if (src == dest) continue;
      // unreachable destination
      if (_agents._[src].dists[dest] == USHRT_MAX) continue;
      uint64_t traffic = get_traffic_volume(src, dest);
      // too little flow
      if (!traffic) continue;
      // add to the end user traffic at the destination
      bupc_atomicU64_fetchadd_relaxed(&_agents._[dest].end_user_traffic, (uint64_t)traffic);
      // starting bad traffic
      double bad_traffic = _agents._[i].wickedness * traffic;
      // egress filtering will reduce bad traffic immediately
      if ( _config.intrvn_drop_egress > 0 && is_active_intrvn(i)) {
	traffic -= (bad_traffic * _config.intrvn_drop_egress);
	bad_traffic *= (1.0 - _config.intrvn_drop_egress);
      }
      // add traffic to source
      bupc_atomicU64_fetchadd_relaxed(&_agents._[i].traffic, traffic);
      // record if some fraction of bad traffic has already been dropped
      int dropped_bad_traffic = 0;
      // to ensure that we don't get stuck following an excessively long path. For error checking.
      int path_length = 0;
      // follow shortest path from src to dest
      while (src != dest) {
        int next_hop = _agents._[dest].paths[src];
	// checking for errors
        if (next_hop >= _agents.num) {
          FAIL("trying to propagate flow from %d to %d (%d) via non-existent neighbor, %d\n", 
               i, dest, src, next_hop);
        }
        if (next_hop == src) {
          FAIL("trying to propagate flow back to oneself %d, %d\n", i, next_hop);
        }
        path_length++;
	if (path_length > 30) {
	  printf("WARNING: path length from %d to %d is > 30\n", i, dest);
	  break;
	}
        if (src == dest && path_length > _agents._[i].dists[dest]) {
          printf("Too many Hops %d from %d to %d, reachable %.2f\n", 
                 path_length - _agents._[i].dists[dest], i, dest, _agents._[dest].reachable);
        }
	// check for blacklisting
	if (_agents._[i].wickedness >= _config.wkdns_to_block && 
	    _agents._[i].num_links < _config.too_big_too_block &&
	    is_active_intrvn(next_hop)) {
	  uint64_t lost_bad_traffic = traffic * _agents._[i].wickedness;
	  bupc_atomicU64_fetchadd_relaxed(&_agents._[dest].lost_bad_traffic, lost_bad_traffic);
	  uint64_t lost_good_traffic = traffic - lost_bad_traffic;
	  bupc_atomicU64_fetchadd_relaxed(&_agents._[dest].lost_good_traffic, lost_good_traffic);
	  traffic = 0;
	  break;
	}
	// Filtering transit traffic. Only need to do it once if it happens
	if (_config.intrvn_drop_transit && !dropped_bad_traffic && is_active_intrvn(next_hop)) {
	  dropped_bad_traffic = 1;
	  traffic -= _config.intrvn_drop_transit * bad_traffic;
	  bad_traffic *= (1.0 - _config.intrvn_drop_transit);
	}
	if (next_hop == dest) {
	  // check to see if ingress traffic will be dropped at the destination
	  if (_config.intrvn_drop_ingress && !dropped_bad_traffic && is_active_intrvn(next_hop)) {
	    traffic -= _config.intrvn_drop_ingress * bad_traffic;
	    bad_traffic *= (1.0 - _config.intrvn_drop_ingress);
	  }
	  // only the dest receives the bad traffic
	  if (get_double_rnd(0, 1.0) <= bad_traffic - trunc(bad_traffic)) bad_traffic++;
	  bupc_atomicU64_fetchadd_relaxed(&_agents._[next_hop].bad_traffic, (uint64_t)bad_traffic);
	}
	bupc_atomicU64_fetchadd_relaxed(&_agents._[next_hop].traffic, traffic);
        src = next_hop;
      } // while (src != dest)
    }
  }
}

uint64_t get_traffic_volume(int src, int dest) {
  // uses something like the gravity model, but distance is determined by distance on the AS graph
  int d = _agents._[src].dists[dest];
  double t = (double)_agents._[src].pop_sh * (double)_agents._[dest].pop_sh / (d * d);
  if (!_config.full_traffic) t /= ((double)MAX_POP * (double)MAX_POP);
  if (get_double_rnd(0, 1.0) <= t - trunc(t)) t++;
  return t;
}

uint64_t get_int_traffic_volume(int i) {
  // uses something like the gravity model, but distance is determined by distance on the AS graph
  double t = (double)_agents._[i].pop_sh * (double)_agents._[i].pop_sh;
  if (!_config.full_traffic) t /= ((double)MAX_POP * (double)MAX_POP);
  if (get_double_rnd(0, 1.0) <= t - trunc(t)) t++;
  return t;
}

int is_active_intrvn(int agent_i) {
  //  if (_agents._[agent_i].num_links >= _config.intrvn_large_isp_deg && _agents._[agent_i].active_intrvns) return 1;
  if (_agents._[agent_i].active_intrvns == ACTIVE) return 1;
  return 0;
}

// this does a breadth-first search of the graph to find all pairs shortest paths
void compute_graph_paths() {
  if ((_sim_iters % GRAPH_EVAL_FREQ) && _agents.num < _config.max_agents) return;
  upc_forall (int agent_i = 0; agent_i < _agents.num; agent_i++; agent_i) {
    shared agent_t* agent = get_agent_ptr(agent_i, __LINE__);
    agent->reachable = 1;
    agent->av_dist = 0;
    agent->max_dist = 0;
    for (int i = 0; i < _agents.num; i++) agent->dists[i] = agent->paths[i] = USHRT_MAX;
    agent->dists[agent_i] = 0;
    agent->paths[agent_i] = agent_i;
    // enqueue the starting vertex 
    if (!enqueue(agent->dist_q, agent_i)) FAIL("no more queue space\n", NULL);
    int num_alt_paths = 0;
    while (!queue_is_empty(agent->dist_q)) {
      int v = dequeue(agent->dist_q);
      for (int i = 0; i < _agents._[v].num_links; i++) {
        int w = _agents._[v].links_list[i];
        int dist = agent->dists[v] + 1;
	if (dist < agent->dists[w]) {
          agent->dists[w] = dist;
          agent->paths[w] = v;
          agent->av_dist += dist;
          agent->reachable++;
          if (agent->max_dist < dist) agent->max_dist = dist;
	  if (!enqueue(agent->dist_q, w)) FAIL("no more queue space\n", NULL);
          num_alt_paths++;
	} else if (dist == agent->dists[w]) {
	  num_alt_paths++;
          // there is some probability that we will use this one for the shortest path
	  if (get_double_rnd(0.0, 1.0) < 1.0 / num_alt_paths) agent->paths[w] = v;
	}
      }
    }
    agent->av_dist /= _agents.num;
    agent->reachable /= _agents.num;
    if (agent->reachable < 1.0) {
      printf("Thread %d WARNING: agent %d only reaches %.3f other agents\n", 
	     MYTHREAD, agent->id, agent->reachable);
    }
  }
}

uint64_t get_traffic_from_deg(int deg) {
  if (deg == 0) return 0;
  double pfp_attach_exp = 0.1;
  double traf = pow((double)deg, 1.0 + pfp_attach_exp * log10(deg));
  if (get_double_rnd(0, 1.0) <= traf - trunc(traf)) traf++;
  return traf;
}

double get_av_deg() {
  return (double)_agents.num_links / _agents.num;
}

double get_locs_covered() {
  return (double)_locs.num_occupied / _locs.num;
}

double get_av_agent_size() {
  double tot_sz = 0;
  for (int i = 0; i < _agents.num; i++) tot_sz +=  _agents._[i].num_locs;
  return tot_sz / _agents.num;
}

int get_max_agent_size() {
  int max_sz = 0;
  for (int i = 0; i < _agents.num; i++) {
    if (_agents._[i].num_locs > max_sz) max_sz = _agents._[i].num_locs;
  }
  return max_sz;
}

double get_av_agent_pop_sh() {
  double tot_pop = 0;
  for (int i = 0; i < _agents.num; i++) tot_pop += _agents._[i].pop_sh;
  return tot_pop / _agents.num;
}

double get_global_bad_traffic() {
  double tot_bad_traf = 0;
  for (int i = 0; i < _agents.num; i++) tot_bad_traf += _agents._[i].bad_traffic;
  return tot_bad_traf;
}

double get_global_lost_good_traffic() {
  double tot_lost_traf = 0;
  for (int i = 0; i < _agents.num; i++) tot_lost_traf += _agents._[i].lost_good_traffic;
  return tot_lost_traf;
}

double get_global_lost_bad_traffic() {
  double tot_lost_traf = 0;
  for (int i = 0; i < _agents.num; i++) tot_lost_traf += _agents._[i].lost_bad_traffic;
  return tot_lost_traf;
}

double get_global_traffic() {
  double tot_traf = 0;
  for (int i = 0; i < _agents.num; i++) tot_traf += _agents._[i].traffic;
  return tot_traf;
}

double get_global_end_user_traffic() {
  double tot_traf = 0;
  for (int i = 0; i < _agents.num; i++) {
    tot_traf += _agents._[i].end_user_traffic;
  }
  return tot_traf;
}

int get_tot_actives() {
  int tot_actives = 0;
  for (int i = 0; i < _agents.num; i++) {
    if (_agents._[i].num_links >= _config.intrvn_large_isp_deg && _agents._[i].active_intrvns) tot_actives++;
  }
  return tot_actives;
}

shared loc_t* get_loc_ptr(int loc_i) {
  if (loc_i < 0 || loc_i >= _locs.num) FAIL("loc_i %d out of range\n", loc_i);
  return &_locs._[loc_i];
}

shared agent_t* get_agent_ptr(int agent_i, int line_num) {
  if (agent_i < 0 || agent_i >= _agents.num) FAIL("agent_i %d out of range from line %d\n", agent_i, line_num);
  return &_agents._[agent_i];
}

void print_stats(double start_t, double *prev_t) {
  static char buf[100] = "";
  double curr_t = _get_current_time() - start_t;
  if (!_config.snapshot_fname && _sim_iters > _config.output_interval) {
    double t_gradient = (log(curr_t) - log(*prev_t)) / 
      (log(_sim_iters) - log(_sim_iters - _config.output_interval));
    double t_intersect = log(curr_t) - t_gradient * log(_sim_iters);
    double eta = pow(NAT_LOG_E, t_gradient * log(_config.max_agents) + t_intersect);
    time_t t_secs = round(start_t + eta);
    struct tm *tmp = localtime(&t_secs);
    strftime(buf, sizeof(buf), "%T", tmp);
  }
  *prev_t = curr_t;
  double av_agent_size = get_av_agent_size();
  int max_agent_size = get_max_agent_size();
  double end_user_traffic = get_global_end_user_traffic();
  double lost_good_traffic = get_global_lost_good_traffic();
  double lost_bad_traffic = get_global_lost_bad_traffic();
  double bad_traffic = get_global_bad_traffic();
  double bad_traffic_frac = 0;
  if (end_user_traffic - lost_good_traffic - lost_bad_traffic > 0) {
    bad_traffic_frac = bad_traffic / (end_user_traffic - lost_good_traffic - lost_bad_traffic);
  }
  double lost_good_traffic_frac = 0;
  if (end_user_traffic - bad_traffic - lost_bad_traffic > 0) {
    lost_good_traffic_frac = lost_good_traffic / (end_user_traffic - bad_traffic - lost_bad_traffic);
  }
  mprintf(13, "%6d", _agents.num, 
	  "%7.3f", get_av_deg(), 
	  "%6d", _agents.max_deg,
	  "%7.2f", get_locs_covered(), 
	  "%8.2f", av_agent_size, 
	  "%8d", max_agent_size,
	  "%10.1e", get_av_agent_pop_sh(), 
	  "%10.1e", get_global_traffic() / _agents.num,
	  "%10.1e", end_user_traffic / _agents.num,
	  "%8.3f", bad_traffic_frac,
	  "%8.3f", lost_good_traffic_frac,
	  "%8.2f", _agents.av_wickedness,
	  "%8d", get_tot_actives());
  if (!_config.snapshot_fname && _sim_iters > _config.output_interval) printf("%10s", buf);
  printf("\n");
  fflush(stdout);
}

void compute_aprobs() {
  // repeatedly sample using geo sel to determine the attachment probs
  int num_samples = 1000000 / THREADS;
  for (int i = 0; i < num_samples; i++) {
    shared loc_t* loc = get_loc_ptr(get_occupied_rnd_loc());
    shared agent_t* agent = get_agent_ptr(loc->agent_list[get_int_rnd(loc->num_agents)], __LINE__);
    bupc_atomicI32_fetchadd_relaxed(&agent->selection_count, 1);
  }
  upc_barrier;
  SERIAL {
    int tot_selection_count = 0;
    for (int i = 0; i < _agents.num; i++) tot_selection_count += _agents._[i].selection_count;
    FILE* f = fopen("aprobs.dat", "w+");
    print_config(&_config, '#', f);
    mfprintf(f, 11, "#%6s", "agent", "%7s", "deg", "%7s", "ext", "%11s", "aprob", "%6s", "av d", 
	     "%6s", "max d", "%11s", "traf", "%10s", "popsh", "%10s", "pop", "%8s", "rchble", 
	     "%8s", "reaches", "%11s\n");
    for (int i = 0; i < _agents.num; i++) {
      shared agent_t* agent = get_agent_ptr(i, __LINE__);
      mfprintf(f, 11, "%7d", i, "%7d", agent->num_links, "%7d", agent->num_locs, 
	       "%11.2e", (double)agent->selection_count / (double)tot_selection_count, 
	       "%6.2f", agent->av_dist, "%6.0f", agent->max_dist, 
	       "%11.2e", (double)agent->traffic, "%10.3f", (double)agent->pop_sh / (double)_agents.num, 
	       "%10.0f", (double)agent->tot_pop, "%8.3f", 1.0, "%8.3f\n", 1.0);
    }
    fclose(f);
  }
}


