#ifndef __ASIM_H
#define __ASIM_H

#include <upc_relaxed.h>
#include <upc_collective.h>

typedef struct {
  int id;
  uint64_t traffic;                   // total traffic routed by this agent
  uint64_t end_user_traffic;          // total traffic received by end-users, including any dropped good or bad traf
  double funds;                       // the total funds held by this agent currently
  int num_links; 
  int tot_pop;                        // how much population an agent serves across all locations
  double pop_sh;                      // must be divided by _config.max_agents to get the actual value
  int num_locs;                       // number of locations this agent exists at
  int is_stub;

  // for verification
  int birth_time;                     // the iteration this agent was born
  int iters;                          // the number of iterations this agent has existed
  double reachable;                   // frac of nodes that reach this one
  
  // statistics
  double av_dist;
  double max_dist;
  int32_t selection_count;            // used for determining selection probabilities.
                                      // used simply for post-hoc analysis of algorithm behavior
  // for cyber-security related issues
  uint64_t bad_traffic;               // malicious traffic received by this agent for its end users 
  double wickedness;                  // each agent has a certain level of compromised hosts
  int active_intrvns;                 // true if this agent will actively intervene if it is big enough
  uint64_t lost_good_traffic;         // good traffic that is dropped
  uint64_t lost_bad_traffic;          // bad traffic that is dropped

  // arrays for neighbors and distances
  // these are all of fixed size as a workaround for a upc alloc bug. 
  shared unsigned short* links;       // array of all agents, indicating which link to this one
  shared unsigned short* links_list;  // list of all agents this one links to
  shared unsigned short* locs;        // array of all locs, marking the ones this agent exists at
  shared unsigned short* locs_list;   // list of all the locs this agent is at
  shared unsigned short* dists;       // distances to other agents
  shared unsigned short* paths;       // shortest paths to other agents
  shared queue_t* dist_q;             // q for storing other agents when calculating distances
} agent_t;

typedef struct {
  int id;
  int32_t num_agents;   // number of agents at this loc
  int32_t pop;       // population at this loc
  shared unsigned short* agent_list;   // list of all agents at this loc
} loc_t;

shared struct {
  shared agent_t* _;
  int num;              // number of agents currently in existence
  int num_links;        // number of links currently existing
  int32_t max_deg;      // current max deg of any agent
  double av_wickedness; // wickedness averaged over all agents
  int32_t num_actives;  // the total number of actively intervening agents
} _agents;

shared struct {
  shared loc_t* _;
  int num;              // total number of locs
  int32_t num_occupied; // number of locs occupied by agents
  int tot_pop;          // population summed over all locs
  int32_t occupied_pop; // population summed over all occupied locs
  int min_pop;          // min allowable pop per loc
  int max_pop;          // max allowable pop per loc
} _locs;

double verify();
void print_timers();
void print_agents();
void print_locs();
void print_links();
void print_dists();

void save_snapshot();
void restore_snapshot();
 
double get_av_deg();
double get_locs_covered();
double get_av_agent_size();
int get_max_agent_size();
double get_av_agent_pop_sh();
int is_active_intrvn(int agent_i);

shared loc_t* get_loc_ptr(int loc_i);
shared agent_t* get_agent_ptr(int agent_i, int line_num);

// globals always prefixed with '_'
shared config_t _config;   
shared int _sim_iters;

#endif


