#include <getopt.h>
#include <stdlib.h>
#include <stdio.h>
#include <upc_relaxed.h>
#include "config.h"

static struct option lopts[] = {
  {"out_interval", 1, 0, 'i'},
  {"random_seed", 1, 0, 'r'},
  {"max_agents", 1, 0, 'n'},
  {"pop_distr_exp", 1, 0, 'p'},
  {"av_degree", 1, 0, 'k'},
  {"extent_cost", 1, 0, 'x'},  
  {"base_income", 1, 0, 'b'},
  {"wickedness", 1, 0, 'w'},
  {"load_snapshot", 1, 0, 's'},
  {"full_traffic", 0, 0, 'f'},
  {"intrvn_drop_ingress", 1, 0, 'g'},
  {"intrvn_drop_egress", 1, 0, 'e'},
  {"intrvn_drop_transit", 1, 0, 't'},
  {"intrvn_large_isp_deg", 1, 0, 'd'},
  {"intrvn_active_prob", 1, 0, 'a'},
  {"recovery_rate", 1, 0, 'R'},
  {"infct_rate", 1, 0, 'I'},
  {"wkdns_to_block", 1, 0, 'W'},
  {"too_big_too_block", 1, 0, 'B'},
  {"verbose_flags", 1, 0, 'v'},
  {"help", 0, 0, 'h'},
  {0, 0, 0, 0}
};


const char* opts_help[] = {
  "output interval in iterations", 
  "random seed (reproducible for given thread count)",
  "desired number of agents", 
  "exponent for population distribution", 
  "desired average degree",
  "extension cost",
  "base income",
  "wickedness level", 
  "file name to load snapshot from",
  "full traffic calculation (much slower)",
  "fraction of bad ingress traffic to drop",
  "fraction of bad egress traffic to drop",
  "fraction of bad transit traffic to drop",
  "min deg for lange ISP intervetions",
  "prob of a large ISP intervening",
  "rate at which infected end-users get clean",
  "rate of infection spread by bad traffic",
  "wckdns above which all traffic is dropped",
  "don't block traffic above this degree",
  "",
  "this help"};

void load_config(int argc, char** argv, shared config_t* config) {
  char verbose_flag_help[5000] = "verbose: ";
  char buf[5000];
  int vflags_len = sizeof(VERBOSE_FLAGS) / sizeof(verbose_flag_t);
  for (int i = 0; i < vflags_len; i++) {
    sprintf(buf, "%c=%s,", VERBOSE_FLAGS[i].flag, VERBOSE_FLAGS[i].name);
    strcat(verbose_flag_help, buf);
    if (i == 4) strcat(verbose_flag_help, "\n                              ");
  }
  // set defaults
  // these are set to give reasonable outputs
  config->output_interval = 1000;
  config->rseed = 29;
  config->pop_distr_exp = -1;
  config->av_degree = 4.202;
  config->max_agents = 10000;
  config->ext_cost = 1.5;
  config->base_income = 5;
  config->wickedness = 0.1;
  config->snapshot_fname = NULL;
  config->full_traffic = 0;
  config->intrvn_drop_ingress = 0;
  config->intrvn_drop_egress = 0;
  config->intrvn_drop_transit = 0;
  config->intrvn_large_isp_deg = MAX_DEG + 1;
  config->intrvn_active_prob = 0;
  config->recovery_rate = 0;
  config->infct_rate = 0;
  config->wkdns_to_block = 1;
  config->too_big_too_block = 0;
  config->verbose_flags = 0;
  char verbose_str[5000] = "";
  // get configs
  char optstr[100] = "+";
  for (int i = 0, j = 1; lopts[i].name; i++) {
    optstr[j++] = lopts[i].val;
    if (lopts[i].has_arg) optstr[j++] = ':';
  }
  int opt;
  while (1) {
    int option_index = 0;
    opt = getopt_long(argc, argv, optstr, lopts, &option_index);
    if (opt == -1) break;
    switch (opt) {
    case 'i': config->output_interval = atoi(optarg); break;
    case 'r': config->rseed = atoi(optarg); break;
    case 'n': config->max_agents = atoi(optarg); break;
    case 'p': config->pop_distr_exp = atof(optarg); break;
    case 'k': config->av_degree = atof(optarg); break; 
    case 'x': config->ext_cost = atof(optarg); break;
    case 'b': config->base_income = atoi(optarg); break;
    case 'w': config->wickedness = atof(optarg); break;
    case 's': 
      config->snapshot_fname = upc_global_alloc(255, sizeof(char));
      strcpy((char*)config->snapshot_fname, optarg);
      break;
    case 'f': config->full_traffic = 1; break;
    case 'g': config->intrvn_drop_ingress = atof(optarg); break;
    case 'e': config->intrvn_drop_egress = atof(optarg); break;
    case 't': config->intrvn_drop_transit = atof(optarg); break;
    case 'd': config->intrvn_large_isp_deg = atoi(optarg); break;
    case 'a': config->intrvn_active_prob = atof(optarg); break;
    case 'R': config->recovery_rate = atof(optarg); break;
    case 'I': config->infct_rate = atof(optarg); break;
    case 'W': config->wkdns_to_block = atof(optarg); break;
    case 'B': config->too_big_too_block = atoi(optarg); break;
    case 'v': 
      for (int i = 0; i < strlen(optarg); i++) {
        for (int v = 0; v < vflags_len; v++) {
          if (optarg[i] == VERBOSE_FLAGS[v].flag) {
            config->verbose_flags |= VERBOSE_FLAGS[v].index;
            sprintf(buf, "%c", VERBOSE_FLAGS[v].flag);
            strcat(verbose_str, buf);
          }
        }
      }
      break;
    default: opt = 'h';
    }
    if (opt == 'h') break;
  }
  if (opt == 'h') {
    // print help
    printf("Usage:  %s options\nWhere options are:\n", argv[0]);
    int i = 0;
    while (lopts[i].name) {
      if (!strcmp(lopts[i].name, "verbose_flags")) {
        printf("   -%c --%-20s: %s\n", lopts[i].val, lopts[i].name, 
               verbose_flag_help);
      } else {
        printf("   -%c --%-20s: %s\n", lopts[i].val, lopts[i].name, opts_help[i]);
      }
      i++;
    }
    exit(EXIT_FAILURE);
  }
  print_config(config, '#', stdout);
  fprintf(stdout, "%c  -v%s\n", '#', verbose_str);
}

void print_config(shared config_t* config, char comment, FILE* f) {
  int i = 0;
  fprintf(f, "%c  -%c %-50s %d\n", comment, lopts[i].val, opts_help[i], config->output_interval); i++;
  fprintf(f, "%c  -%c %-50s %d\n", comment, lopts[i].val, opts_help[i], config->rseed); i++;
  fprintf(f, "%c  -%c %-50s %d\n", comment, lopts[i].val, opts_help[i], config->max_agents); i++;
  fprintf(f, "%c  -%c %-50s %.3f\n", comment, lopts[i].val, opts_help[i], config->pop_distr_exp); i++;
  fprintf(f, "%c  -%c %-50s %.3f\n", comment, lopts[i].val, opts_help[i], config->av_degree); i++;  
  fprintf(f, "%c  -%c %-50s %.3f\n", comment, lopts[i].val, opts_help[i], config->ext_cost); i++;
  fprintf(f, "%c  -%c %-50s %d\n", comment, lopts[i].val, opts_help[i], config->base_income); i++;
  fprintf(f, "%c  -%c %-50s %.2f\n", comment, lopts[i].val, opts_help[i], config->wickedness); i++;
  fprintf(f, "%c  -%c %-50s %s\n", comment, lopts[i].val, opts_help[i], (char*)config->snapshot_fname); i++;
  fprintf(f, "%c  -%c %-50s %d\n", comment, lopts[i].val, opts_help[i], config->full_traffic); i++;
  fprintf(f, "%c  -%c %-50s %.3f\n", comment, lopts[i].val, opts_help[i], config->intrvn_drop_ingress); i++;
  fprintf(f, "%c  -%c %-50s %.3f\n", comment, lopts[i].val, opts_help[i], config->intrvn_drop_egress); i++;
  fprintf(f, "%c  -%c %-50s %.3f\n", comment, lopts[i].val, opts_help[i], config->intrvn_drop_transit); i++;
  fprintf(f, "%c  -%c %-50s %d\n", comment, lopts[i].val, opts_help[i], config->intrvn_large_isp_deg); i++;
  fprintf(f, "%c  -%c %-50s %.3f\n", comment, lopts[i].val, opts_help[i], config->intrvn_active_prob); i++;
  fprintf(f, "%c  -%c %-50s %.3f\n", comment, lopts[i].val, opts_help[i], config->recovery_rate); i++;
  fprintf(f, "%c  -%c %-50s %.3f\n", comment, lopts[i].val, opts_help[i], config->infct_rate); i++;
  fprintf(f, "%c  -%c %-50s %.3f\n", comment, lopts[i].val, opts_help[i], config->wkdns_to_block); i++;
  fprintf(f, "%c  -%c %-50s %d\n", comment, lopts[i].val, opts_help[i], config->too_big_too_block); i++;
  //  fprintf(f, "%c     %-50s %d\n", comment, "link strategy", LINK_STRATEGY);
}



