#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <upc_relaxed.h>
#include "utils.h"

static unsigned int _rnd_seed = 29;
static double _elapsed_time[MAX_TIMERS];
static double _start_time[MAX_TIMERS];

void init_rnd(unsigned int rseed) {
  _rnd_seed = rseed;
}

int get_int_rnd(int range) {
  if (range <= 0) {
    FAIL("Range for get_int_rnd <= 0: %d\n", range);
  }
  int rnd_index = ((double)range * rand_r(&_rnd_seed)) / RAND_MAX;
  if (rnd_index > range - 1 || rnd_index < 0) {
    FAIL("Error in random number generation index out of range, %d > %d\n", 
           rnd_index, range - 1);
  }
  return rnd_index;
}

double get_double_rnd(double min, double max) {
  return ((double)rand_r(&_rnd_seed)) / (double)RAND_MAX * (max - min) + min;
}

// returns an index into the list, randomly selected according to the elementweight,
// given by the func elem_weight. 
// the func elem_weight should renturn a positive weight for all elements to be considered, 
// and a zero weight for all elements to be skipped
// 'param' is passed on to the func, for internal use
int get_list_rnd(int num, int tot_weight, shared void* list, 
		 double (*elem_weight)(int, shared void*, void*), void* param) {
  if (tot_weight <= 0) return -1;
  int rnd_pos = get_int_rnd(tot_weight);
  for (int i = 0; i < num; i++) {
    rnd_pos -= elem_weight(i, list, param);
    if (rnd_pos < 0) return i;
  }
  return -1;
}

double _get_current_time(void) {
  //  double t = bupc_ticks_to_us(bupc_ticks_now()) * 1.0e-6;            
  struct timespec ts;
  syscall(__NR_clock_gettime, CLOCK_REALTIME, &ts);
  double t = (double)ts.tv_sec + (double)ts.tv_nsec * 1e-9;
  return t;
}

void clear_timers() {
  for (int i = 0; i < MAX_TIMERS; i++) timer_clear(i);
}

void timer_clear(int n) {
  _elapsed_time[n] = 0.0;
}

void timer_start(int n) {
  _start_time[n] = _get_current_time();
}

void timer_stop(int n) {
  double t, now;
  now = _get_current_time();
  t = now - _start_time[n];
  _elapsed_time[n] += t;
}

double timer_read(int n) {
  return(_elapsed_time[n]);
}

int init_queue(shared queue_t* q, int q_len) {
  if (q_len == 0) return 0;
  q->entries = upc_global_alloc(q_len, sizeof(int));
  q->front = q->back = 0;
  q->len = q_len;
  return 1;
}

void print_queue(shared queue_t* q) {
  printf("(%d, %d) ", q->front, q->back);
  if (q->front == q->back) {
    printf("\n");
    return;
  }
  for (int k = 0; k < q->len; k++) {
    int ki = k + q->front;
    if (ki == q->len) {
      ki == 0;
      if (q->back = q->len) break;
    }
    printf("%d ", q->entries[ki]);
    if (ki == q->back - 1) break;
  }
  printf("\n");
}

void clear_queue(shared queue_t* q) {
  q->front = 0;
  q->back = 0;
  //  for (int i = 0; i < q->len; i++) q->in_q[i] = 0;
}

// We use this instead of strstr because that generates warnings in UPC
int in_str(const char* str, const char* substr) {
  int sub_i = 0;
  int started = 0;
  for (int i = 0; i < strlen(str); i++) {
    if (!started) {
      if (str[i] == substr[sub_i] ) started = 1;
    }
    if (started) {
      if (str[i] != substr[sub_i]) return 0;
      sub_i++;
      if (sub_i == strlen(substr)) return 1;
    }
  }
  return 0;

}

void mfprintf(FILE* f, int num_args, ...) {
  va_list ap;
  va_start(ap, num_args);
  for (int i = 0; i < num_args; i++) {
    char* fmt = va_arg(ap, char*);
    // now determine what type the arg should be
    if (in_str(fmt, "d")) fprintf(f, fmt, va_arg(ap, int));
    if (in_str(fmt, "s")) fprintf(f, fmt, va_arg(ap, char*));
    if (in_str(fmt, "f")) fprintf(f, fmt, va_arg(ap, double));
    if (in_str(fmt, "e")) fprintf(f, fmt, va_arg(ap, double));
  }
  va_end(ap);
}

void heap_init(heap_t* heap, int max_size) {
  heap->elems = calloc(max_size, sizeof(heap_elem_t));
  heap->in_heap = calloc(max_size, sizeof(unsigned short));
  heap->size = 0;
  heap->max_size = max_size;
}

int heap_insert(heap_t* heap, int key, int val) {
  if (heap_is_full(heap)) return 0;
  int i = ++heap->size;
  while (heap->elems[i / 2].val > val) {
    cp_heap_elem(&heap->elems[i], &heap->elems[i / 2]);
    i /= 2;
  }
  heap->elems[i].key = key;
  heap->elems[i].val = val;
  heap->in_heap[key] = 1;
  return 1;
}

int heap_remove(heap_t* heap, heap_elem_t* elem) {
  if (heap_is_empty(heap)) return 0;
  heap_elem_t min_elem, last_elem;
  cp_heap_elem(&min_elem, &heap->elems[1]);
  cp_heap_elem(&last_elem, &heap->elems[heap->size]);
  heap->size--;
  int child = 0, i;
  for (i = 1; i * 2 <= heap->size; i = child) {
    child = i * 2;
    if ((child != heap->size) && (heap->elems[child + 1].val < heap->elems[child].val)) child++;
    if (last_elem.val > heap->elems[child].val) cp_heap_elem(&heap->elems[i], &heap->elems[child]);
    else break;
  }
  cp_heap_elem(&heap->elems[i], &last_elem);
  heap->in_heap[min_elem.key] = 0;
  cp_heap_elem(elem, &min_elem);
  return 1;
}

void heap_clear(heap_t* heap) {
  heap->size = 0;
  for (int i = 0; i < heap->max_size; i++) heap->in_heap[i] = 0;
}

int heap_is_empty(heap_t* heap) {
  return !heap->size;
}

int heap_is_full(heap_t* heap) {
  return heap->size >= heap->max_size;
}

int heap_change(heap_t* heap, int key, int delta) {
  return 0;
}

void cp_heap_elem(heap_elem_t* dest, const heap_elem_t* src) {
  dest->key = src->key;
  dest->val = src->val;
}

static int cmp_int(const void* a, const void* b) {
  int ai = *(int *)a;
  int bi = *(int *)b;
  return (ai > bi ? 1 : (ai < bi ? -1 : 0));
}

void par_sort(shared int* A, int num) {
  if (THREADS == 1) {
    int* my_A = calloc(num, sizeof(int));
    for (int i = 0; i < num; i++) my_A[i] = A[i];
    qsort(my_A, num, sizeof(int), cmp_int);
    for (int i = 0; i < num; i++) A[i] = my_A[i];
    return;
  }
  shared int* split_pts = upc_all_alloc(THREADS - 1, sizeof(int));
  shared int* split_sizes = upc_all_alloc(THREADS, sizeof(int));
  split_sizes[MYTHREAD] = 0;
  upc_barrier;
  SERIAL {
    int* sp = calloc(THREADS - 1, sizeof(int));
    for (int i = 0; i < THREADS - 1; i++) sp[i] = A[get_int_rnd(num)];
    qsort(sp, THREADS - 1, sizeof(int), cmp_int);
    for (int i = 0; i < THREADS - 1; i++) {
      split_pts[i] = sp[i];
      printf("thread %d has split pt %d\n", i, split_pts[i]);
    }
    free(sp);
  }
  upc_barrier;
  int* split_array = calloc(num, sizeof(int));
  for (int i = 0; i < num; i++) {
    if (MYTHREAD == THREADS - 1) {
      if (A[i] > split_pts[MYTHREAD - 1]) split_array[split_sizes[MYTHREAD]++] = A[i];
    } else if (MYTHREAD == 0) {
      if (A[i] <= split_pts[MYTHREAD]) split_array[split_sizes[MYTHREAD]++] = A[i];
    } else {
      if (A[i] <= split_pts[MYTHREAD] && A[i] > split_pts[MYTHREAD - 1]) {
	  split_array[split_sizes[MYTHREAD]++] = A[i];
      }
    }
  }
  printf("thread %d is qsorting %d elems\n", MYTHREAD, split_sizes[MYTHREAD]);
  qsort(split_array, split_sizes[MYTHREAD], sizeof(int), cmp_int);
  upc_barrier;
  int start_pt = 0;
  for (int i = 0; i < MYTHREAD; i++) start_pt += split_sizes[i];
  for (int i = 0; i < split_sizes[MYTHREAD]; i++) A[i + start_pt] = split_array[i];
  free(split_array);
  upc_barrier;
  SERIAL {
    upc_free(split_pts);
    upc_free(split_sizes);
  }
}

#if 0

#define NUM 100

int main() {
  shared int* A = upc_all_alloc(NUM, sizeof(int));
  upc_barrier;
  SERIAL {
    for (int i = 0; i < NUM; i++) A[i] = get_int_rnd(1000);
  }
  upc_barrier;
  timer_clear(0);
  timer_start(0);
  par_sort(A, NUM);
  timer_stop(0);
  SERIAL { 
    //    for (int i = 0; i < NUM; i++) printf("%d %d\n", i, A[i]);
    for (int i = 0; i < NUM - 1; i++) {
      if (A[i] > A[i + 1]) printf("Error in sorting at pos %d, %d, %d > %d\n", i, i + 1, A[i], A[i + 1]);
    }
  }
  SERIAL printf("took %.3fs\n", timer_read(0));
  return 0;
}

#endif


#if 0
// for testing lists

#define NUM_T 10000000
#define NUM_A 9

int a1[] = {0,1,2,6,0,4,0,3,0};
shared int* a;

/*
double get_weight(int i, shared void* list, void* p) {
  return ((shared int*)list)[i];
}
*/


GET_LIST_RND(get_rnd_slot(int num, double tot_weight), (a[i]));

int main(int argc, char* argv[]) {
  struct timeval tv;
  gettimeofday(&tv, NULL);
  int rseed = 29;//tv.tv_usec;
  init_rnd(rseed + MYTHREAD);
  SERIAL {
    a = upc_global_alloc(NUM_A, sizeof(int));
    for (int i = 0; i < NUM_A; i++) a[i] = a1[i];
    int counts[NUM_A];
    int tot_counts = 0;
    for (int i = 0; i < NUM_A; i++) {
      tot_counts += a[i];
      counts[i] = 0;
    }
    printf("tot counts %d\n", tot_counts);
    //    for (int i = 0; i < NUM_T; i++) counts[get_list_rnd(NUM_A, tot_counts, a, get_weight, NULL)]++;
    for (int i = 0; i < NUM_T; i++) counts[get_rnd_slot(NUM_A, tot_counts)]++;
    for (int i = 0; i < NUM_A; i++) {
      printf("%d %.3f %.3f\n", i, (double)a[i] / tot_counts, (double)counts[i] / NUM_T);
    }
  }
}

#endif
