#ifndef _UTILS_H
#define _UTILS_H

#include <upc_relaxed.h>
#include <sys/syscall.h>
#include <stdio.h>

#define SERIAL if (MYTHREAD == 0)
#define BITS_PER_ULONG (sizeof(unsigned long) * 8)
#define BIT_INDEX(X) (X - (BITS_PER_ULONG * (int)(X / BITS_PER_ULONG)))

#define MAX_TIMERS 20
char* _timer_names[MAX_TIMERS];

#define CREATE_TIMER(t) static int t = __COUNTER__

#define PHASE(func)\
  CREATE_TIMER(TIMER_##func);\
  CREATE_TIMER(TIMER_BARRIER_##func);\
  if (!_timer_names[TIMER_##func]) {\
    _timer_names[TIMER_##func] = #func;                                   \
    _timer_names[TIMER_BARRIER_##func] = "(barriers)"; \
    timer_clear(TIMER_##func);                                            \
    timer_clear(TIMER_BARRIER_##func);                                    \
  }\
  timer_start(TIMER_##func);\
  func();			     \
  timer_start(TIMER_BARRIER_##func); \
  upc_barrier; \
  timer_stop(TIMER_##func); \
  timer_stop(TIMER_BARRIER_##func)

#define FAIL(fmt, ...)							\
  do {printf("%s:%d [%d] FAILURE: " fmt, __FILE__, __LINE__, MYTHREAD, __VA_ARGS__); \
    upc_global_exit(-1);} while (0);

#define LOCK(lock) if (THREADS > 1) upc_lock(lock)
#define UNLOCK(lock) if (THREADS > 1) upc_unlock(lock)

void init_rnd(unsigned int rseed);
int get_int_rnd(int range);
double get_double_rnd(double min, double max);
int get_list_rnd(int num, int tot_weight, shared void* list, 
		 double (*elem_weight)(int, shared void*, void*), void* param);
double _get_current_time(void);
void timer_clear(int n);
void timer_start(int n);
void timer_stop(int n);
double timer_read(int n);
void mfprintf(FILE* f, int num_args, ...);
#define mprintf(num_args, ...) mfprintf(stdout, num_args, __VA_ARGS__)

// this does pretty much the same thing as the function above, but it is more efficient, and
// results in much neater code, since the elem_weight function is integrated with the declaration
#define GET_LIST_RND(func, elem_weight) int func {			\
  if (tot_weight <= 0) return -1;					\
    int rnd_pos = get_int_rnd(tot_weight);				\
    for (int i = 0; i < num; i++) {					\
      rnd_pos -= elem_weight;						\
      if (rnd_pos < 0) return i;					\
    }									\
    return -1; }


typedef struct {
  shared int* entries;
  int front, back, len;
} queue_t;

int init_queue(shared queue_t* q, int q_len);
void print_queue(shared queue_t* q);
void clear_queue(shared queue_t* q);

inline int enqueue(shared queue_t* q, int v) {
  q->entries[q->back++] = v;
  if (q->back == q->len) q->back = 0;
  if (q->back == q->front) return 0;
  //  printf("enqueued %d\n", v);
  //  print_queue(q);
  return 1;
}

inline int dequeue(shared queue_t* q) {
  int v = q->entries[q->front++];
  if (q->front == q->len) q->front = 0;
  //  printf("dequeued %d\n", v);
  //  print_queue(q);
  return v;
}

inline int queue_is_empty(shared queue_t* q) {
  return q->front == q->back;
}

typedef struct {
  shared int* entries;
  int num;
  int len;
} stack_t;

inline void init_stack(shared stack_t* s, int len) {
  s->entries = upc_global_alloc(len, sizeof(int));
  s->len = len;
  s->num = 0;
}

inline int stack_is_empty(shared stack_t* s) {
  return s->num;
}

inline int pop_stack(shared stack_t* s) {
  if (s->num == 0) return -1;
  return s->entries[--s->num];
}

inline int push_stack(shared stack_t* s, int v) {
  if (s->num == s->len) return -1;
  s->entries[s->num++] = v;
  return 0;
}

inline void clear_stack(shared stack_t* s) {
  s->num = 0;
}

inline int stack_index(shared stack_t* s, unsigned i) {
  if (i >= s->num) return -1;
  return s->entries[i];
}

typedef struct {
  int key, val;
} heap_elem_t;

typedef struct {
  unsigned int max_size;
  unsigned int size;
  heap_elem_t* elems;
  unsigned short* in_heap;
} heap_t;

void cp_heap_elem(heap_elem_t* dest, const heap_elem_t* src);
void heap_init(heap_t* heap, int max_size);
int heap_insert(heap_t* heap, int key, int val);
int heap_remove(heap_t* heap, heap_elem_t* elem);
int heap_change(heap_t* heap, int key, int delta);
void heap_clear(heap_t* heap);
int heap_is_empty(heap_t* heap);
int heap_is_full(heap_t* heap);

#endif
