#include <math.h>
#include "utils.h"
#include "config.h"
#include "asim.h"

shared int32_t num_occupied = 0;
shared int32_t occupied_pop = 0;

double verify() {
  CREATE_TIMER(VERIFY_TIMER);
  timer_clear(VERIFY_TIMER);
  timer_start(VERIFY_TIMER);
  shared int* tot_num_links = upc_all_alloc(THREADS, sizeof(int));
  tot_num_links[MYTHREAD] = 0;
  upc_forall(int i = 0; i < _agents.num; i++; i) {
    shared agent_t* agent = &_agents._[i];
    // 1. verify iterations
    if (agent->iters != _sim_iters - agent->birth_time) {
      printf("Verification *FAILED*: agent %d has iterations (%d) != age (%d)\n", 
	     i, agent->iters, _sim_iters - agent->birth_time);
    }
    // 2. verify pop
    int count_num_locs = 0;
    int count_pop = 0;
    for (int l = 0; l < _locs.num; l++) {
      shared loc_t* loc = &_locs._[l];
      if (agent->locs[l]) {
        count_num_locs++;
        count_pop += loc->pop;
      }
    }     
    if (count_pop != agent->tot_pop) {
      printf("Verification *FAILED*: Agent %d pop is %d but across locations it's %d\n", 
	     i, agent->tot_pop, count_pop);
    }
    // 2. verify agent locs
    for (int l = 0; l < _locs.num; l++) {
      shared loc_t* loc = &_locs._[l];
      if (agent->locs[l]) {
	int found = 0;
	for (int al = 0; al < loc->num_agents; al++) {
	  if (loc->agent_list[al] == agent->id) {
	    found = 1;
	    break;
	  }
	}
	if (!found) {
	  printf("Verification *FAILED*: Agent %d says it's at loc %d, but this is not recorded\n", 
		 i, l);
	}
      }
    }
    // 3. verify locations
    if (count_num_locs != agent->num_locs) {
      printf("Verification *FAILED*: Agent %d says it's at %d locs but it's actually at %d locs\n", 
             i, agent->num_locs, count_num_locs);
    }
    // 4. verify links
    int agent_num_links = 0;
    for (int j = 0; j < _agents.num; j++) {
      shared agent_t* agent_2 = &_agents._[j];
      if (agent->links[j] && !agent_2->links[i]) {
          printf("Verification *FAILED*: Agent %d is linked to %d, but not vice versa\n", i, j);
      } else if (agent->links[j]) {
	tot_num_links[MYTHREAD]++;
	agent_num_links++;
      }
    }
    for (int j = 0; j < agent->num_links; j++) {
      int link_i = agent->links_list[j];
      if (!agent->links[link_i] || !_agents._[link_i].links[i]) {
	printf("Verification *FAILED*: Agent %d thinks it is linked to %d\n", i, link_i);
      }
    }
    if (agent_num_links != agent->num_links) {
      printf("Verification *FAILED*: Agent %d says it has %d links, but it actually has %d\n", 
             i, agent->num_links, agent_num_links);
    }
  }
  upc_barrier;
  // 5. verify total and occupied pop
  upc_forall(int i = 0; i < _locs.num; i++; i) {
    if (_locs._[i].num_agents) {
      bupc_atomicI32_fetchadd_relaxed(&num_occupied, 1);
      bupc_atomicI32_fetchadd_relaxed(&occupied_pop, _locs._[i].pop);
    }
  }
  upc_barrier;
  SERIAL {
    if (num_occupied != _locs.num_occupied) {
      printf("Verification *FAILED*: There are %d occupied locs but only %d are recorded\n",
	     num_occupied, _locs.num_occupied);
    }
    if (occupied_pop != _locs.occupied_pop) {
      printf("Verification *FAILED*: The occupied pop is %d but only %d recorded\n",
	     occupied_pop, _locs.occupied_pop);
    }
  }
  upc_barrier;
  SERIAL {
    // 6. verify agent update counts and total link counts
    int num_links = 0;
    for (int i = 0; i < THREADS; i++) num_links += tot_num_links[i];
    if (num_links != _agents.num_links) {
      printf("Verification *FAILED*: Threads count %d links but the total is given as %d\n", 
             num_links, _agents.num_links);
    } 
  }
  timer_stop(VERIFY_TIMER);
  upc_barrier;
  double time_elapsed = timer_read(VERIFY_TIMER);
  SERIAL printf("# Verification in %.2fs\n", time_elapsed);
  return time_elapsed;
}
